"""Add user.cache_dir

Revision ID: 385aa8c18489
Revises: 47072e0106ce
Create Date: 2023-04-06 10:13:43.234442

"""
import sqlalchemy as sa
import sqlmodel
from alembic import op


# revision identifiers, used by Alembic.
revision = "385aa8c18489"
down_revision = "47072e0106ce"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "user_oauth",
        sa.Column(
            "cache_dir", sqlmodel.sql.sqltypes.AutoString(), nullable=True
        ),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("user_oauth", "cache_dir")
    # ### end Alembic commands ###



# Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
# University of Zurich
#
# Original authors:
# Jacopo Nespolo <jacopo.nespolo@exact-lab.it>
# Tommaso Comparin <tommaso.comparin@exact-lab.it>
#
# This file is part of Fractal and was originally developed by eXact lab S.r.l.
# <exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
# Institute for Biomedical Research and Pelkmans Lab from the University of
# Zurich.
"""
Runner backend subsystem root

This module is the single entry point to the runner backend subsystem. Other
subystems should only import this module and not its submodules or the
individual backends.
"""
import os
from pathlib import Path
from typing import Optional

from ... import __VERSION__
from ...config import get_settings
from ...logger import set_logger
from ...syringe import Inject
from ...utils import get_timestamp
from ..db import DB
from ..models import ApplyWorkflow
from ..models import Dataset
from ..models import JobStatusType
from ..models import Workflow
from ._local import process_workflow as local_process_workflow
from .common import auto_output_dataset  # noqa: F401
from .common import close_job_logger
from .common import JobExecutionError
from .common import TaskExecutionError
from .common import validate_workflow_compatibility  # noqa: F401


_backends = {}
_backend_errors: dict[str, Exception] = {}

_backends["local"] = local_process_workflow

try:
    from ._slurm import process_workflow as slurm_process_workflow

    _backends["slurm"] = slurm_process_workflow
except ModuleNotFoundError as e:
    _backend_errors["slurm"] = e


def get_process_workflow():
    settings = Inject(get_settings)
    try:
        process_workflow = _backends[settings.FRACTAL_RUNNER_BACKEND]
    except KeyError:
        raise _backend_errors.get(
            settings.FRACTAL_RUNNER_BACKEND,
            RuntimeError(
                "Unknown error during collection of backend "
                f"`{settings.FRACTAL_RUNNER_BACKEND}`"
            ),
        )
    return process_workflow


async def submit_workflow(
    *,
    workflow_id: int,
    input_dataset_id: int,
    output_dataset_id: int,
    job_id: int,
    worker_init: Optional[str] = None,
    slurm_user: Optional[str] = None,
    user_cache_dir: Optional[str] = None,
) -> None:
    """
    Prepares a workflow and applies it to a dataset

    This function wraps the process_workflow one, which is different for each
    backend (e.g. local or slurm backend).

    Args:
        workflow_id:
            ID of the workflow being applied
        input_dataset_id
            Input dataset ID
        output_dataset_id:
            ID of the destination dataset of the workflow. If not provided,
            overwriting of the input dataset is implied and an error is raised
            if the dataset is in read only mode.
        job_id:
            Id of the job record which stores the state for the current
            workflow application.
        worker_init:
            Custom executor parameters that get parsed before the execution of
            each task.
        user_cache_dir:
            Cache directory (namely a path where the user can write); for the
            slurm backend, this is used as a base directory for
            `job.working_dir_user`.
        slurm_user:
            The username to impersonate for the workflow execution, for the
            slurm backend.
    """

    db_sync = next(DB.get_sync_db())

    job: ApplyWorkflow = db_sync.get(ApplyWorkflow, job_id)
    if not job:
        raise ValueError("Cannot fetch job from database")
    input_dataset: Dataset = db_sync.get(Dataset, input_dataset_id)
    if not input_dataset:
        raise ValueError("Cannot fetch input_dataset from database")
    output_dataset: Dataset = db_sync.get(Dataset, output_dataset_id)
    if not output_dataset:
        raise ValueError("Cannot fetch output_dataset from database")
    workflow: Workflow = db_sync.get(Workflow, workflow_id)
    if not workflow:
        raise ValueError("Cannot fetch workflow from database")

    # Select backend
    settings = Inject(get_settings)
    FRACTAL_RUNNER_BACKEND = settings.FRACTAL_RUNNER_BACKEND
    process_workflow = get_process_workflow()

    # Prepare some of process_workflow arguments
    input_paths = input_dataset.paths
    output_path = output_dataset.paths[0]
    workflow_id = workflow.id

    # Define and create server-side working folder
    project_id = workflow.project_id
    WORKFLOW_DIR = (
        settings.FRACTAL_RUNNER_WORKING_BASE_DIR  # type: ignore
        / f"proj_{project_id:07d}_wf_{workflow_id:07d}_job_{job_id:07d}"
    ).resolve()
    if not WORKFLOW_DIR.exists():
        original_umask = os.umask(0)
        WORKFLOW_DIR.mkdir(parents=True, mode=0o755)
        os.umask(original_umask)

    # Define and create user-side working folder, if needed
    if FRACTAL_RUNNER_BACKEND == "local":
        WORKFLOW_DIR_USER = WORKFLOW_DIR
    elif FRACTAL_RUNNER_BACKEND == "slurm":
        timestamp_string = get_timestamp().strftime("%Y%m%d_%H%M%S")

        from ._slurm._subprocess_run_as_user import _mkdir_as_user

        WORKFLOW_DIR_USER = (
            Path(user_cache_dir) / f"{timestamp_string}_{WORKFLOW_DIR.name}"
        ).resolve()
        _mkdir_as_user(folder=str(WORKFLOW_DIR_USER), user=slurm_user)
    else:
        raise ValueError(f"{FRACTAL_RUNNER_BACKEND=} not supported")

    # Update db
    job.working_dir = WORKFLOW_DIR.as_posix()
    job.working_dir_user = WORKFLOW_DIR_USER.as_posix()
    job.status = JobStatusType.RUNNING
    db_sync.merge(job)
    db_sync.commit()

    # Write logs
    logger_name = f"WF{workflow_id}_job{job_id}"
    log_file_path = WORKFLOW_DIR / "workflow.log"
    logger = set_logger(
        logger_name=logger_name,
        log_file_path=log_file_path,
    )
    logger.info(
        f'Start execution of workflow "{workflow.name}"; '
        f"more logs at {str(log_file_path)}"
    )
    logger.debug(f"fractal_server.__VERSION__: {__VERSION__}")
    logger.debug(f"FRACTAL_RUNNER_BACKEND: {FRACTAL_RUNNER_BACKEND}")
    logger.debug(f"slurm_user: {slurm_user}")
    logger.debug(f"worker_init: {worker_init}")
    logger.debug(f"input metadata: {input_dataset.meta}")
    logger.debug(f"input_paths: {input_paths}")
    logger.debug(f"output_path: {output_path}")
    logger.debug(f"job.id: {job.id}")
    logger.debug(f"job.working_dir: {str(WORKFLOW_DIR)}")
    logger.debug(f"job.workflow_dir_user: {str(WORKFLOW_DIR_USER)}")
    logger.debug(f'START workflow "{workflow.name}"')

    # Note: from the docs, "The Session.close() method does not prevent the
    # Session from being used again"
    # (https://docs.sqlalchemy.org/en/20/orm/session_api.html#sqlalchemy.orm.Session.close)
    db_sync.close()

    try:

        output_dataset.meta = await process_workflow(
            workflow=workflow,
            input_paths=input_paths,
            output_path=output_path,
            input_metadata=input_dataset.meta,
            slurm_user=slurm_user,
            user_cache_dir=user_cache_dir,
            workflow_dir=WORKFLOW_DIR,
            workflow_dir_user=WORKFLOW_DIR_USER,
            logger_name=logger_name,
            worker_init=worker_init,
        )

        logger.debug(f'END workflow "{workflow.name}"')

        db_sync.merge(output_dataset)

        job.status = JobStatusType.DONE
        with log_file_path.open("r") as f:
            logs = f.read()
        job.log = logs
        db_sync.merge(job)

    except TaskExecutionError as e:

        logger.debug(f'FAILED workflow "{workflow.name}", TaskExecutionError.')
        logger.info(f'Workflow "{workflow.name}" failed (TaskExecutionError).')

        job.status = JobStatusType.FAILED

        exception_args_string = "\n".join(e.args)
        job.log = (
            f"TASK ERROR:"
            f"Task id: {e.workflow_task_id} ({e.task_name}), "
            f"{e.workflow_task_order=}\n"
            f"TRACEBACK:\n{exception_args_string}"
        )
        db_sync.merge(job)

    except JobExecutionError as e:

        logger.debug(f'FAILED workflow "{workflow.name}", JobExecutionError.')
        logger.info(f'Workflow "{workflow.name}" failed (JobExecutionError).')

        job.status = JobStatusType.FAILED
        error = e.assemble_error()
        job.log = f"JOB ERROR:\nTRACEBACK:\n{error}"
        db_sync.merge(job)

    except Exception as e:

        logger.debug(f'FAILED workflow "{workflow.name}", unknown error.')
        logger.info(f'Workflow "{workflow.name}" failed (unkwnon error).')

        job.status = JobStatusType.FAILED
        job.log = f"UNKNOWN ERROR\nOriginal error: {str(e)}"
        db_sync.merge(job)

    finally:
        close_job_logger(logger)
        db_sync.commit()
        db_sync.close()



"""Drop project.project_dir

Revision ID: 6dede8d6fd9d
Revises: 385aa8c18489
Create Date: 2023-04-06 10:32:52.972906

"""
import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "6dede8d6fd9d"
down_revision = "385aa8c18489"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("project", "project_dir")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "project", sa.Column("project_dir", sa.VARCHAR(), nullable=False)
    )
    # ### end Alembic commands ###



# Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
# University of Zurich
#
# Original authors:
# Jacopo Nespolo <jacopo.nespolo@exact-lab.it>
#
# This file is part of Fractal and was originally developed by eXact lab S.r.l.
# <exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
# Institute for Biomedical Research and Pelkmans Lab from the University of
# Zurich.
"""
This module takes care of installing tasks so that fractal can execute them

Tasks can be private or public. Private tasks are installed under
`Settings.FRACTAL_TASKS_DIR/{username}`. For public tasks, `username =
.fractal`.
"""
import json
import shutil
import sys
from io import IOBase
from pathlib import Path
from typing import Optional
from typing import Union
from zipfile import ZipFile

from pydantic import root_validator

from ..common.schemas import ManifestV1
from ..common.schemas import TaskCollectPip
from ..common.schemas import TaskCollectStatus
from ..common.schemas import TaskCreate
from ..config import get_settings
from ..logger import get_logger
from ..syringe import Inject
from ..utils import execute_command


FRACTAL_PUBLIC_TASK_SUBDIR = ".fractal"


def get_python_interpreter(version: Optional[str] = None) -> str:
    """
    Return the path to the python interpreter

    Args:
        version: Python version

    Raises:
        ValueError: If the python version requested is not available on the
                    host.

    Returns:
        interpreter: string representing the python executable or its path
    """
    if version:
        interpreter = shutil.which(f"python{version}")
        if not interpreter:
            raise ValueError(
                f"Python version {version} not available on host."
            )
    else:
        interpreter = sys.executable

    return interpreter


def get_absolute_venv_path(venv_path: Path) -> Path:
    """
    Note:
    In Python 3.9 it would be safer to do:

        if venv_path.is_relative_to(settings.FRACTAL_TASKS_DIR):
            package_path = venv_path
        else:
            package_path = settings.FRACTAL_TASKS_DIR / venv_path
    """
    if venv_path.is_absolute():
        package_path = venv_path
    else:
        settings = Inject(get_settings)
        package_path = settings.FRACTAL_TASKS_DIR / venv_path  # type: ignore
    return package_path


def get_collection_path(base: Path) -> Path:
    return base / "collection.json"


def get_log_path(base: Path) -> Path:
    return base / "collection.log"


def get_collection_log(venv_path: Path) -> str:
    package_path = get_absolute_venv_path(venv_path)
    log_path = get_log_path(package_path)
    log = log_path.open().read()
    return log


def get_collection_data(venv_path: Path) -> TaskCollectStatus:
    package_path = get_absolute_venv_path(venv_path)
    collection_path = get_collection_path(package_path)
    with collection_path.open() as f:
        data = json.load(f)
    return TaskCollectStatus(**data)


class _TaskCollectPip(TaskCollectPip):
    """
    Internal TaskCollectPip schema

    The difference with its parent class is that we check if the package
    corresponds to a path in the filesystem.
    """

    package_path: Optional[Path] = None

    @property
    def is_local_package(self) -> bool:
        return bool(self.package_path)

    @root_validator(pre=True)
    def check_local_package(cls, values):
        """
        Checks if package corresponds to an existing path on the filesystem

        In this case, the user is providing directly a package file, rather
        than a remote one from PyPI. We set the `package_path` attribute and
        get the actual package name and version from the package file name.
        """
        if "/" in values["package"]:
            package_path = Path(values["package"])
            if not package_path.is_absolute():
                raise ValueError("Package path must be absolute")
            if package_path.exists():
                values["package_path"] = package_path
                (
                    values["package"],
                    values["version"],
                    *_,
                ) = package_path.name.split("-")
        return values

    @property
    def pip_package_version(self):
        """
        Return pip compatible specification of package and version
        """
        version = f"=={self.version}" if self.version else ""
        return f"{self.package}{version}"

    @property
    def source(self):
        if self.is_local_package:
            return f"pip-local:{self.package_path.name}"
        else:
            return f"pip:{self.pip_package_version}"

    def check(self):
        if not self.version:
            raise ValueError("Version is not set or cannot be determined")


def _package_from_path(wheel_path: Path) -> tuple[str, str]:
    """
    Extract package name and version from package files such as wheel files.
    """
    wheel_filename = wheel_path.name
    package, version, *_rest = wheel_filename.split("-")
    return package, version


def create_package_dir_pip(
    *,
    task_pkg: _TaskCollectPip,
    user: Optional[str] = None,
    create: bool = True,
    **_,  # FIXME remove this catch-all argument
) -> Path:
    settings = Inject(get_settings)
    user = user or FRACTAL_PUBLIC_TASK_SUBDIR

    package_dir = f"{task_pkg.package}{task_pkg.version or ''}"
    venv_path = settings.FRACTAL_TASKS_DIR / user / package_dir  # type: ignore
    # TODO check the access right of the venv_path and subdirs
    if create:
        venv_path.mkdir(exist_ok=False, parents=True)
    return venv_path


async def download_package(
    *,
    task_pkg: _TaskCollectPip,
    dest: Union[str, Path],
):
    """
    Download package to destination
    """
    interpreter = get_python_interpreter(version=task_pkg.python_version)
    pip = f"{interpreter} -m pip"
    cmd = (
        f"{pip} download --no-deps {task_pkg.pip_package_version} "
        f"-d {dest}"
    )
    stdout = await execute_command(command=cmd, cwd=Path("."))
    pkg_file = next(
        line.split()[-1] for line in stdout.split("\n") if "Saved" in line
    )
    return Path(pkg_file)


def inspect_package(path: Path) -> dict:
    """
    Inspect task package for version and manifest

    Args:
        path: Path
            the path in which the package is saved

    Returns:
        version_manifest: A dictionary containing `version`, the version of the
        pacakge, and `manifest`, the Fractal manifest object relative to the
        tasks.
    """
    if "whl" in path.as_posix():
        # it is simply a zip file
        # we can extract the version number from *.dist-info/METADATA
        # and read the fractal manifest from the package content
        with ZipFile(path) as wheel:
            namelist = wheel.namelist()
            metadata = next(
                name for name in namelist if "dist-info/METADATA" in name
            )
            manifest = next(
                name for name in namelist if "__FRACTAL_MANIFEST__" in name
            )

            with wheel.open(metadata) as metadata_fd:
                meta = metadata_fd.read().decode("utf-8")
                version = next(
                    line.split()[-1]
                    for line in meta.splitlines()
                    if line.startswith("Version")
                )

            with wheel.open(manifest) as manifest_fd:
                manifest_obj = read_manifest(manifest_fd)  # type: ignore

    version_manifest = dict(version=version, manifest=manifest_obj)
    return version_manifest


async def create_package_environment_pip(
    *,
    task_pkg: _TaskCollectPip,
    venv_path: Path,
    logger_name: str,
) -> list[TaskCreate]:
    """
    Create environment and install package
    """
    logger = get_logger(logger_name)
    try:
        logger.debug("Creating venv and installing package")

        python_bin, package_root = await _create_venv_install_package(
            path=venv_path,
            task_pkg=task_pkg,
            logger_name=logger_name,
        )

        logger.debug("Loading task manifest")

        task_list = load_manifest(
            package_root=package_root,
            python_bin=python_bin,
            source=task_pkg.source,
        )
        logger.debug("Loaded task manifest")
    except Exception as e:
        logger.error("Task manifest loading failed")
        raise e
    return task_list


def read_manifest(file: Union[Path, IOBase]) -> ManifestV1:
    """
    Read and parse manifest file
    """
    if isinstance(file, IOBase):
        manifest_dict = json.load(file)
    else:
        with file.open("r") as f:
            manifest_dict = json.load(f)

    manifest_version = str(manifest_dict["manifest_version"])
    if manifest_version == "1":
        manifest = ManifestV1(**manifest_dict)
    else:
        raise ValueError("Manifest version {manifest_version=} not supported")

    return manifest


def load_manifest(
    package_root: Path,
    python_bin: Path,
    source: str,
) -> list[TaskCreate]:

    manifest_file = package_root / "__FRACTAL_MANIFEST__.json"
    manifest = read_manifest(manifest_file)

    task_list = []
    for t in manifest.task_list:
        task_executable = package_root / t.executable
        if not task_executable.exists():
            raise FileNotFoundError(
                f"Cannot find executable `{task_executable}` "
                f"for task `{t.name}`"
            )
        cmd = f"{python_bin.as_posix()} {task_executable.as_posix()}"
        this_task = TaskCreate(**t.dict(), command=cmd, source=source)
        task_list.append(this_task)
    return task_list


async def _create_venv_install_package(
    *,
    task_pkg: _TaskCollectPip,
    path: Path,
    logger_name: str,
) -> tuple[Path, Path]:
    """Create venv and install package

    Args:
        path: the directory in which to create the environment
        task_pkg: object containing the different metadata required to install
            the package

    Returns:
        python_bin: path to venv's python interpreter
        package_root: the location of the package manifest
    """
    python_bin = await _init_venv(
        path=path,
        python_version=task_pkg.python_version,
        logger_name=logger_name,
    )
    package_root = await _pip_install(
        venv_path=path, task_pkg=task_pkg, logger_name=logger_name
    )
    return python_bin, package_root


async def _init_venv(
    *,
    path: Path,
    python_version: Optional[str] = None,
    logger_name: str,
) -> Path:
    """
    Set a virtual environment at `path/venv`

    Args:
        path : Path
            path to directory in which to set up the virtual environment
        python_version : default=None
            Python version the virtual environment will be based upon

    Returns:
        python_bin : Path
            path to python interpreter
    """
    interpreter = get_python_interpreter(version=python_version)
    await execute_command(
        cwd=path,
        command=f"{interpreter} -m venv venv",
        logger_name=logger_name,
    )
    return path / "venv/bin/python"


async def _pip_install(
    venv_path: Path,
    task_pkg: _TaskCollectPip,
    logger_name: str,
) -> Path:
    """
    Install package in venv

    Returns:
        package_root : Path
            the location of the package manifest
    """
    pip = venv_path / "venv/bin/pip"

    if task_pkg.is_local_package:
        pip_install_str = task_pkg.package_path.as_posix()  # type: ignore
    else:
        version_string = f"=={task_pkg.version}" if task_pkg.version else ""
        extras = (
            f"[{task_pkg.package_extras}]" if task_pkg.package_extras else ""
        )
        pip_install_str = f"{task_pkg.package}{extras}{version_string}"

    cmd_install = f"{pip} install {pip_install_str}"
    cmd_inspect = f"{pip} show {task_pkg.package}"

    await execute_command(
        cwd=venv_path,
        command=f"{pip} install --upgrade pip",
        logger_name=logger_name,
    )
    await execute_command(
        cwd=venv_path, command=cmd_install, logger_name=logger_name
    )

    # Extract package installation path from `pip show`
    stdout_inspect = await execute_command(
        cwd=venv_path, command=cmd_inspect, logger_name=logger_name
    )

    location = Path(
        next(
            line.split()[-1]
            for line in stdout_inspect.split("\n")
            if line.startswith("Location:")
        )
    )
    package_root = location / task_pkg.package.replace("-", "_")
    if not package_root.exists():
        raise RuntimeError(
            "Could not determine package installation location."
        )
    return package_root
